/*
 * $Id: my_pattern.c,v 1.29 2014/04/30 11:36:57 andy Exp $
 *
 * Pattern-related functions for ng_grep module 
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/syslog.h>
#include <netgraph/ng_message.h>
#include <netgraph/netgraph.h>
#include <netgraph/ng_parse.h>
#include "ng_grep.h"
#include "my_pattern.h"

static int patterns_chain_proc(pat_p, int);
static int pattern_proc(pat_p, u_char*, void**, u_char, int);

extern int pattern_add(pat_p pattern, u_char *text, void **prec, u_char flags){
	return(pattern_proc(pattern, text, prec, flags, MY_PATTERN_CMD_ADD));
}
extern int pattern_del(pat_p pattern, u_char *text, void **prec, u_char flags){
	return(pattern_proc(pattern, text, prec, flags, MY_PATTERN_CMD_DEL));
}
extern int pattern_check(pat_p pattern, u_char *text, void **prec, u_char flags){
	return(pattern_proc(pattern, text, prec, flags, MY_PATTERN_CMD_CHECK));
}

extern int patterns_chain_flush(pat_p pattern){
	return(patterns_chain_proc(pattern, MY_PATTERN_CMD_FLUSH));
}
extern int patterns_chain_free(pat_p pattern){
	return(patterns_chain_proc(pattern, MY_PATTERN_CMD_FREE));
}

static int pattern_proc(pat_p pattern, u_char *text, void **found, u_char flags, int opcode){

	int	error	= 0;
	int	match	= 0;
	pat_p	prev;
	u_char	*p_http	= "http://";
	u_char	*delim;
	u_char	host[NG_GREP_MAX_STR_PAR_LEN + 1]	= "";
	u_char	path[NG_GREP_MAX_STR_PAR_LEN + 1]	= "";
	u_int	host_l;
	u_int	path_l;

//	log(LOG_DEBUG,"ng_grep::pattern_proc proc pattern '%s' cmd '%d' \n", text,opcode);

	if(flags & P_FLAG_URL){
//		log(LOG_DEBUG,"ng_grep::pattern_proc proc as URL \n");
		if(strncasecmp(p_http, text, strlen(p_http))){
			log(LOG_DEBUG,"ng_grep::pattern_proc URL:bad protocol  \n");
			error = EINVAL;
		}else{
			if((delim = strchr(text+strlen(p_http),'/'))){
				host_l = delim-text-strlen(p_http);
				strncpy(host,text+strlen(p_http), host_l);
				path_l = min(strlen(delim), NG_GREP_MAX_STR_PAR_LEN);
				strncpy(path,delim,NG_GREP_MAX_STR_PAR_LEN);
			}else{
				host_l = min(strlen(text+strlen(p_http)), NG_GREP_MAX_STR_PAR_LEN - strlen(p_http));
				strncpy(host,text+strlen(p_http),host_l);
				path_l=0;
			}
			path[path_l]=0;
			host[host_l]=0;

//			log(LOG_DEBUG,"ng_grep::pattern_proc host: '%s', path: '%s'\n", host, path);
		}
	}


	if(strlen(text) && !error){
		prev = pattern;
		while (1){
//			log(LOG_DEBUG,"ng_grep::pattern_proc inspecting current  entry for pattern '%s'\n", text);
			if(flags & P_FLAG_URL ){
//				log(LOG_DEBUG,"ng_grep::pattern_proc processing as URL\n");
				if(pattern->precomp && pattern->flag_url){
					
					if(!strcasecmp(((p_url_p)pattern->precomp)->host, host)){
						if(flags & P_FLAG_I_C){
							if(!strcasecmp(((p_url_p)pattern->precomp)->path, path))match = 1;
						}else{
							if(!strcmp(((p_url_p)pattern->precomp)->path,path))match=1;
						}
					}
				}
			}else{
//				log(LOG_DEBUG,"ng_grep::pattern_proc processing as RAW\n");
				if(pattern->text && !pattern->flag_url){
					if(flags & P_FLAG_I_C){
						if(!strcasecmp(pattern->text,text))match = 1;
					}else{
						if(!strcmp(pattern->text,text))match=1;
					}
				}
			}
			if(match){
				if((!pattern->flag_i_c && (flags & P_FLAG_I_C))
				  || (pattern->flag_i_c && !(flags & P_FLAG_I_C))){match = 0;}
			}
			if(match){
				if((!pattern->flag_sub && (flags & P_FLAG_SUB))
				  || (pattern->flag_sub && !(flags & P_FLAG_SUB))){match = 0;}
			}
			if(match){
//				log(LOG_DEBUG,"ng_grep::pattern_proc pattern '%s' found in chain\n", text);
				if(opcode == MY_PATTERN_CMD_DEL){
//					log(LOG_DEBUG,"ng_grep::pattern_proc pattern '%s' must die!\n", text);
					if(prev == pattern ){
						log(LOG_ERR,"ng_grep::pattern_proc can't delete root pattern");
						error = 129;
					}else{
						prev->next = pattern->next;
						if(pattern->text)free(pattern->text,M_NETGRAPH_GREP);
						if(pattern->precomp){
							if(pattern->flag_url){
//								log(LOG_DEBUG,"ng_grep::pattern_proc processing as URL\n");
								if(((p_url_p)pattern->precomp)->host)free(((p_url_p)pattern->precomp)->host,M_NETGRAPH_GREP);
								if(((p_url_p)pattern->precomp)->path)free(((p_url_p)pattern->precomp)->path,M_NETGRAPH_GREP);
							}
							free(pattern->precomp,M_NETGRAPH_GREP);
						}
						free(pattern,M_NETGRAPH_GREP);
					}
				}else{
					*found=pattern;
				}
				break;
			}else{
				if(pattern->next == NULL){
					if(opcode == MY_PATTERN_CMD_ADD){
//						log(LOG_DEBUG,"ng_grep::pattern_proc adding new entry for pattern '%s'\n", text);
						pattern->next = malloc(sizeof(*pattern->next), M_NETGRAPH_GREP, M_ZERO|M_NOWAIT);
						if(flags & P_FLAG_I_C)pattern->flag_i_c = 1;
						if(flags & P_FLAG_SUB)pattern->flag_sub = 1;
						pattern->meet_c = 0;
						if(flags & P_FLAG_URL){
//							log(LOG_DEBUG,"ng_grep::pattern_proc processing as URL\n");
							pattern->precomp = malloc(sizeof(struct precomp_url), M_NETGRAPH_GREP, M_ZERO|M_NOWAIT);				// !!!!!!
							if((pattern->next && pattern->precomp)
							  && (((p_url_p)pattern->precomp)->host=malloc(host_l + 1, M_NETGRAPH_GREP, M_ZERO|M_NOWAIT))
							  && (((p_url_p)pattern->precomp)->path=malloc(path_l + 1, M_NETGRAPH_GREP, M_ZERO|M_NOWAIT))){
								strncpy(((p_url_p)pattern->precomp)->host, host, host_l);
								strncpy(((p_url_p)pattern->precomp)->path, path, path_l);
								((p_url_p)pattern->precomp)->host[host_l]=0;
								((p_url_p)pattern->precomp)->path[path_l]=0;
								((p_url_p)pattern->precomp)->host_l = host_l;
								((p_url_p)pattern->precomp)->path_l = path_l;
								pattern->flag_url = 1;
								*found = pattern;
							}else{
//								log(LOG_ERR,"ng_grep::pattern_proc can't allocate memory for pattern '%s'\n", text);
								if(((p_url_p)pattern->precomp)->host)free(((p_url_p)pattern->precomp)->host, M_NETGRAPH_GREP);
								if(((p_url_p)pattern->precomp)->path)free(((p_url_p)pattern->precomp)->path, M_NETGRAPH_GREP);
								if(pattern->precomp)free(pattern->precomp, M_NETGRAPH_GREP);
								if(pattern->next)free(pattern->next, M_NETGRAPH_GREP);
								error = 128;
							}
						}else{
//							log(LOG_DEBUG,"ng_grep::pattern_proc processing as RAW\n");
							pattern->text = malloc(strlen(text)+1, M_NETGRAPH_GREP, M_ZERO|M_NOWAIT);
							if(pattern->next && pattern->text){
								strcpy(pattern->text,text);
								pattern->length = strlen(text);
							}else{
//								log(LOG_ERR,"ng_grep::pattern_proc can't allocate memory for pattern '%s'\n", text);
								if(pattern->next)free(pattern->next, M_NETGRAPH_GREP);
								if(pattern->text)free(pattern->text, M_NETGRAPH_GREP);
								error = 128;
							}
						}
					}else{
						error = ENOENT;
					}
					break;
				}
			}
			prev = pattern;
			pattern = pattern->next;
//			log(LOG_DEBUG,"ng_grep::pattern_proc ----- and again -----\n");
		}
	}else{
		error = EINVAL;
	}
	return(error);
};

extern pat_p patterns_chain_init(){
	pat_p	root_pattern;
	root_pattern = malloc(sizeof(*root_pattern), M_NETGRAPH_GREP, M_ZERO|M_WAITOK);
	root_pattern->next = malloc(sizeof(*root_pattern), M_NETGRAPH_GREP, M_ZERO|M_WAITOK);
	return(root_pattern);
}

static int patterns_chain_proc(pat_p pattern, int opcode){
	int	error	= 0;
	pat_p	next;
	
	if(opcode == MY_PATTERN_CMD_FLUSH){
		if(pattern->text){
			free(pattern->text,M_NETGRAPH_GREP);
			pattern->text = NULL;
		}
		if(pattern->precomp){
			if(((p_url_p)pattern->precomp)->host)free(((p_url_p)pattern->precomp)->host, M_NETGRAPH_GREP);
			if(((p_url_p)pattern->precomp)->path)free(((p_url_p)pattern->precomp)->path, M_NETGRAPH_GREP);
			free(pattern->precomp, M_NETGRAPH_GREP);
			pattern->precomp = NULL;
		}
		if(pattern->next->text){
			free(pattern->next->text,M_NETGRAPH_GREP);
			pattern->next->text = NULL;
		}
		if(pattern->precomp){
			if(((p_url_p)pattern->next->precomp)->host)free(((p_url_p)pattern->next->precomp)->host, M_NETGRAPH_GREP);
			if(((p_url_p)pattern->next->precomp)->path)free(((p_url_p)pattern->next->precomp)->path, M_NETGRAPH_GREP);
			free(pattern->next->precomp, M_NETGRAPH_GREP);
			pattern->next->precomp = NULL;
		}
		next = pattern->next->next;
		pattern->next->next = NULL;
	}else{
		next = pattern;
	}
	
	while(next){
		pattern = next;	
		next = pattern->next;
		if(pattern->text)free(pattern->text,M_NETGRAPH_GREP);
		if(pattern->precomp){
			if(((p_url_p)pattern->precomp)->host)free(((p_url_p)pattern->precomp)->host, M_NETGRAPH_GREP);
			if(((p_url_p)pattern->precomp)->path)free(((p_url_p)pattern->precomp)->path, M_NETGRAPH_GREP);
			free(pattern->precomp, M_NETGRAPH_GREP);
		}
		free(pattern,M_NETGRAPH_GREP);
	}
	return(error);
}

extern int check_incom_string(pat_p pattern, char *string, u_int string_len){
	int	error		= ENOENT;
	u_char	http_ok;
	u_int	c_str;
	u_int	token_p		= 0;
	u_char	flag_delim	= 1;
	u_char	host_found	= 0;
	u_char	path_found	= 0;
	u_int	in_host;
	u_int	in_path;

	struct token in_token[MY_P_MAX_TOKENS];

	if(!string_len){return(error);}

//	log(LOG_DEBUG,"ng_grep::check_incom_string data: '%s' \n", string);

	for(c_str = 0; c_str <= string_len; c_str++){
			if(string[c_str] == 0x20 || string[c_str] == 0x0a || string[c_str] == 0x0d){
				if(!flag_delim){
					flag_delim=1;
					if(token_p){in_token[token_p - 1].leng = c_str - in_token[token_p - 1].leng;}
				}
			}else{
				if(flag_delim){
					in_token[token_p].leng = c_str;
					in_token[token_p].text = &(string[c_str]);
					flag_delim = 0;
					if(++token_p == MY_P_MAX_TOKENS)break;
				}
			}
	}

	for(token_p = 0; (token_p < MY_P_MAX_TOKENS - 1) && !(http_ok = (host_found && path_found)); token_p++){
		if(!path_found){
			if(in_token[token_p].leng == 3 && !strncasecmp(in_token[token_p].text,"GET",3)){
//				log(LOG_DEBUG,"ng_grep::check_incom_string path: path found! index '%d' \n", token_p);
				in_path = token_p + 1;
				path_found = 1;
				continue;
			}
		}
		if(!path_found){
			if(in_token[token_p].leng == 4 && !strncasecmp(in_token[token_p].text,"POST",4)){
//				log(LOG_DEBUG,"ng_grep::check_incom_string path: path found! index '%d' \n", token_p);
				in_path = token_p + 1;
				path_found = 1;
				continue;
			}
		}
		if(!host_found){
			if(in_token[token_p].leng == 5 && !strncasecmp(in_token[token_p].text,"Host:",5)){
//				log(LOG_DEBUG,"ng_grep::check_incom_string path: host found! index '%d' \n", token_p);
				in_host = token_p + 1;
				host_found = 1;
				continue;
			}
		}
	}

//	log(LOG_DEBUG,"ng_grep::check_incom_string path: '%d' host '%d' path_found '%d' host_found '%d' \n", in_path, in_host, path_found, host_found);
//	log(LOG_DEBUG,"ng_grep::check_incom_string path: http_ok: '%d' \n", http_ok);

	while(pattern){
		if(pattern->text && pattern->length){
//			log(LOG_DEBUG,"ng_grep::check_incom_string compare as RAW\n");
			if(pattern->flag_i_c){
				if(!strncasecmp(string,pattern->text,pattern->length)){
//					log(LOG_DEBUG,"ng_grep::check_incom_string *** MEET *** without case\n");
					error = 0;
					pattern->meet_c++;
					break;
				}
			}else{
				if(!strncmp(string,pattern->text,pattern->length)){
//					log(LOG_DEBUG,"ng_grep::check_incom_string *** MEET *** with case\n");
					error = 0;
					pattern->meet_c++;
					break;
				}
			}
		}
		if(pattern->precomp && http_ok){
//			log(LOG_DEBUG,"ng_grep::check_incom_string compare as URL with 'sub' flag='%d'\n",pattern->flag_sub);
//			log(LOG_DEBUG,"ng_grep::check_incom_string path leng '%d', sample path leng '%d'\n", in_token[in_path].leng,((p_url_p)pattern->precomp)->path_l);
			if((((p_url_p)pattern->precomp)->host_l == in_token[in_host].leng)
			  && ((((p_url_p)pattern->precomp)->path_l == in_token[in_path].leng) || (pattern->flag_sub && (((p_url_p)pattern->precomp)->path_l < in_token[in_path].leng)))
			  && !strncasecmp(((p_url_p)pattern->precomp)->host, in_token[in_host].text,((p_url_p)pattern->precomp)->path_l)){
				if(pattern->flag_i_c){
			  		if(!strncasecmp(((p_url_p)pattern->precomp)->path, in_token[in_path].text,((p_url_p)pattern->precomp)->path_l)){
//						log(LOG_DEBUG,"ng_grep::check_incom_string *** MEET *** without case\n");
						error = 0;
						pattern->meet_c++;
						break;
					}
				}else{
			  		if(!strncmp(((p_url_p)pattern->precomp)->path, in_token[in_path].text,((p_url_p)pattern->precomp)->path_l)){
//						log(LOG_DEBUG,"ng_grep::check_incom_string *** MEET *** with case\n");
						error = 0;
						pattern->meet_c++;
						break;
					}
				}
			}
		}
		pattern=pattern->next;
	}
	return(error);
}
